<?php

declare(strict_types=1);

namespace Drupal\Tests\date_filter\Kernel;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\KernelTests\KernelTestBase;
use Drupal\Tests\node\Traits\NodeCreationTrait;
use Drupal\views\Views;

/**
 * Tests date and datetime fields filtering.
 *
 * @group date_filter
 */
final class DateFilterTest extends KernelTestBase {

  use NodeCreationTrait;

  // Time interval for created nodes. Since the initial value of one of the
  // filters is > 1 month (see views.view.date_filter_test), one week will
  // be appropriate.
  private const TIME_INTERVAL = 604_800;

  private const NODE_TYPE = 'dates_test';

  private const TEST_VIEW = 'date_filter_test';
  private const TEST_VIEW_DISPLAY = 'page_1';

  private const TEST_DATA = [
    [
      'created' => '2022-01-01 10:00:00',
      'field_date' => '2022-01-01',
    ],
    [
      'created' => '2022-01-02 11:00:00',
      'field_date' => '2022-02-14',
    ],
    [
      'created' => '2022-02-10 08:00:00',
      'field_date' => '2022-01-02',
    ],
    [
      'created' => '2022-02-11 08:30:00',
      'field_date' => '2022-02-03',
    ],
    [
      'created' => '2022-02-12 08:45:00',
      'field_date' => '2022-02-27',
    ],
  ];

  /**
   * The time service.
   */
  private TimeInterface $time;

  /**
   * The date formatter service.
   */
  private DateFormatterInterface $dateFormatter;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'user',
    'node',
    'field',
    'views',
    // The datetime module is not a general dependency but should be included
    // for a comprehensive test of all supported field types.
    'datetime',
    'date_filter',
    'date_filter_test',
    'system',
  ];

  /**
   * Test nodes data.
   *
   * Includes titles, created timestamps and field_date values.
   *
   * @var mixed[]|null
   */
  protected ?array $testNodesData;

  /**
   * {@inheritdoc}
   */
  public function setUp(): void {
    parent::setUp();

    $this->installEntitySchema('user');
    $this->installEntitySchema('node');
    $this->installSchema('node', 'node_access');
    $this->installSchema('system', 'sequences');

    $this->installConfig([
      'system',
      'date_filter_test',
    ]);

    // Get time and VBO view data services.
    $this->time = $this->container->get('datetime.time');
    $this->dateFormatter = $this->container->get('date.formatter');

    // Create test nodes with different created timestamps and
    // field_date values.
    foreach (self::TEST_DATA as $delta => $node_data) {
      $node_data['created'] = DrupalDateTime::createFromFormat('Y-m-d H:i:s', $node_data['created'])->format('U');

      $node = $this->createNode([
        'type' => self::NODE_TYPE,
        'title' => \sprintf('test node "%d"', $delta),
        'created' => $node_data['created'],
        'field_date' => $node_data['field_date'],
      ]);
      $this->testNodesData[$node->id()] = $node_data;
    }
  }

  /**
   * Tests the initial value of the filter.
   */
  public function testDateFilterInitial(): void {
    $view = Views::getView(self::TEST_VIEW);
    $view->setDisplay(self::TEST_VIEW_DISPLAY);
    $view->execute();

    // Check if the view filters were set to correct dates.
    // Allow 5 second threshold for view execution time.
    $actual_created = \strtotime(implode(' ', $view->exposed_data['created']));
    $expected = \strtotime('-1 month', $this->time->getCurrentTime());
    $condition =
      $expected >= $actual_created &&
      $expected <= $actual_created + 5;
    $this->assertTrue($condition, \sprintf('Created filter actual %d not equal to expected %d', $actual_created, $expected));
  }

  /**
   * Tests the relation operation access.
   *
   * @param int|null $created_filter_value
   *   Value of the node created filter.
   * @param array|null $date_field_filter_value
   *   Value of the node date field filter.
   * @param string $timezone
   *   The test case time zone.
   *
   * @dataProvider filterDataProvider
   */
  public function testDateFilter(?int $created_filter_value, ?array $date_field_filter_value, string $timezone): void {

    // Set timezone.
    $this->config('system.date')
      ->set('timezone.user.configurable', 0)
      ->set('timezone.default', $timezone)
      ->save();

    // Determine what should show and what not first.
    $expected = [
      // -1 month is the view default for created filter.
      'created' => $created_filter_value ?? \strtotime('-1 month', $this->time->getCurrentTime()),
      'field_date' => NULL,
    ];
    if ($date_field_filter_value !== NULL) {
      $expected['field_date'] = [
        'min' => $date_field_filter_value['min'] === NULL ? NULL : \strtotime($date_field_filter_value['min'] . ' 00:00:00'),
        'max' => $date_field_filter_value['max'] === NULL ? NULL : \strtotime($date_field_filter_value['max'] . ' 23:59:59'),
      ];
    }
    $should_show = [];
    foreach ($this->testNodesData as $node_id => $data) {
      $should_show[$node_id] = TRUE;
      $field_date_ts = \strtotime($data['field_date'] . ' 12:00:00');

      // Check all conditions.
      if (
        $expected['created'] !== NULL &&
        $data['created'] <= $expected['created']
      ) {
        $should_show[$node_id] = FALSE;
        continue;
      }
      if ($expected['field_date'] === NULL) {
        continue;
      }
      if (
        $expected['field_date']['min'] !== NULL &&
        $field_date_ts <= $expected['field_date']['min']
      ) {
        $should_show[$node_id] = FALSE;
        continue;
      }
      if (
        $expected['field_date']['max'] !== NULL &&
        $field_date_ts >= $expected['field_date']['max']
      ) {
        $should_show[$node_id] = FALSE;
        continue;
      }
    }

    $view = Views::getView(self::TEST_VIEW);
    $view->setDisplay(self::TEST_VIEW_DISPLAY);

    $exposed_input = [];
    if ($created_filter_value !== NULL) {
      $exposed_input['created'] = [
        'date' => $this->dateFormatter->format($created_filter_value, 'custom', 'Y-m-d'),
        'time' => $this->dateFormatter->format($created_filter_value, 'custom', 'H:i:s'),
      ];
    }
    if ($date_field_filter_value !== NULL) {
      $exposed_input['field_date_value'] = [];
      foreach (['min', 'max'] as $type) {
        if ($date_field_filter_value[$type] !== NULL) {
          $exposed_input['field_date_value'][$type] = [
            'date' => $date_field_filter_value[$type],
          ];
        }
      }
    }

    if (\count($exposed_input) !== 0) {
      $view->setExposedInput($exposed_input);
    }

    // Just in case.
    $view->setItemsPerPage(0);
    $view->setCurrentPage(0);
    $view->setOffset(0);

    // Get result entity IDs and assert if they should be present in
    // the result set.
    $view->execute();

    $result_ids = [];
    foreach ($view->result as $row) {
      $node_id = $row->_entity->id();
      $result_ids[$node_id] = $node_id;
    }

    // Now we know what should show and what not, time for assertions.
    foreach ($should_show as $node_id => $result) {
      if ($result) {
        $this->assertArrayHasKey($node_id, $result_ids, \sprintf(
          'Node %d (created on %d, date field %s) should show in results.',
          $node_id,
          $this->testNodesData[$node_id]['created'],
          $this->testNodesData[$node_id]['field_date']
        ));
      }
      else {
        $this->assertArrayNotHasKey($node_id, $result_ids, \sprintf(
          'Node %d (created on %d, date field %s) should not show in results.',
          $node_id,
          $this->testNodesData[$node_id]['created'],
          $this->testNodesData[$node_id]['field_date']
        ));
      }
    }

  }

  /**
   * Data provider method.
   *
   * @return mixed[]
   *   Test cases data.
   */
  public function filterDataProvider(): array {
    $cases = [];

    foreach ([
      // UTC-12, no DST.
      'Pacific/Kwajalein',
      // UTC-7, no DST.
      'America/Phoenix',
      // UTC.
      'UTC',
      // UTC+5:30, no DST.
      'Asia/Kolkata',
      // UTC+13, no DST.
      'Pacific/Tongatapu',
    ] as $timezone) {
      $cases[] = [
        // This case will eventually not yield any results (default created
        // filter value -1 month from current date), but will verify if
        // this value is set correctly.
        'created_filter_value' => NULL,
        'date_field_filter_value' => NULL,
        'timezone' => $timezone,
      ];
      $cases[] = [
        'created_filter_value' => \strtotime('2022-01-02 10:30:00'),
        'date_field_filter_value' => NULL,
        'timezone' => $timezone,
      ];
      $cases[] = [
        'created_filter_value' => \strtotime('2020-01-01 10:30:00'),
        'date_field_filter_value' => [
          'min' => '2022-02-15',
          'max' => NULL,
        ],
        'timezone' => $timezone,
      ];
      $cases[] = [
        'created_filter_value' => \strtotime('2020-01-01 10:30:00'),
        'date_field_filter_value' => [
          'min' => '2022-01-10',
          'max' => '2022-02-03',
        ],
        'timezone' => $timezone,
      ];
      $cases[] = [
        'created_filter_value' => \strtotime('2020-01-01 10:30:00'),
        'date_field_filter_value' => [
          'min' => NULL,
          'max' => '2022-02-03',
        ],
        'timezone' => $timezone,
      ];

      // @todo Add one more datetime field (date and time) and include
      // test cases for that field as well.
    }

    return $cases;
  }

}
